package ACL2::Demostracion;

#
# Demostración de la raíz del proyecto, con dependencias cumplidas
# Copyright (C) 2005-2008 Antonio García Domínguez
# $Id$
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#

use warnings;
use strict;
use utf8;

use Carp;

use XML::Writer;    # XML
use Cwd 'realpath'; # Manejador de directorios, rutas y nombres de fichero
use File::ShareDir 'dist_dir'; # Directorio de ficheros compartidos - contiene el DTD
use FindBin;        # Obtiene la ruta del script

use ACL2::Dependencias;
use ACL2::Localizacion;
use ACL2::Orden::Fabrica;
use ACL2::Orden;
use ACL2::Lisp::Parser qw(read_lisp_program try_read_lisp_token);
use ACL2::Proyecto;
use ACL2::Tags;
use ACL2::State;
use XML::Writer;
use File::Spec;

my $path_to_dtd;
my $lh;

BEGIN {
    $lh = ACL2::Localizacion->get_handle();

    # We need to detect the path to the DTD file _before_ any directory
    # changes take place. This is mostly useful during development, but
    # doesn't hurt the rest of the time.

    # Path to the system-wide auto folder with the DTD
    my $dtd_carpeta = eval { dist_dir('ACL2-Procesador') } || '.';
    my $dtd_fichero = "script.dtd";

    # List of all possible paths to the DTD
    my @rutasDTD = ("share/${dtd_fichero}",                  # Testing
                    "$FindBin::Bin/share/${dtd_fichero}",    # Dev/PAR
                    "$FindBin::Bin/../share/${dtd_fichero}", # Dev/PAR
                    "$dtd_carpeta/${dtd_fichero}",           # Post-install
                    );

    # Use the first existing path
    foreach my $ruta (@rutasDTD) {
        if (-r $ruta) {
            $path_to_dtd = realpath($ruta);
            last;
        }
    }
    die $lh->maketext("No se halló el DTD: [_1]", join(", ", @rutasDTD))
        if !defined($path_to_dtd);
}

=head1 FUNCTIONS

All public methods and exported functions are listed below.

=head2 new

  Uso: constructor. Does not perform any analysis.
  Arguments: takes a hash with the following keys:

=over 2

=item *

ruta_lisp: Absolute path to the original Lisp script.

=item *

ruta_dem: Absolute path to the .out file with ACL2's output.

=item *

ruta_lisp: Cadena con la ruta completa a la fuente Lisp original, para
posterior referencia.

=item *

deps: ACL2::Dependencias object with the dependencies of the Lisp script.

=back

=cut

sub new {
    my ($clase, %params) = @_;

    my $d = bless {
        "ruta_tmp_lisp" => $params{ruta_tmp_lisp},
        "ruta_lisp" => $params{ruta_lisp},
        "ruta_dem"  => $params{ruta_dem},
        "deps"      => $params{deps},
    }, $clase;

    $d->_inicializar;
    return $d;
}

=head2 escribir_xml

Uso: vuelca la información obtenida a partir del análisis de la
demostración de ACL2 en formato XML.

Parámetros:

=over 2

=item *

writer: Opcional. Indica el objeto XML::Writer sobre el que se debería
producir la salida. Se creará automáticamente a partir de los otros
parámetros si no se pasa.

=item *

destino_xml: Obligatorio si no se ha pasado C<writer>. Indica a dónde
va a parar el documento XML resultante. Puede ser o bien una instancia
de alguna subclase de IO::Handle, o una referencia a una cadena.

=item *

incrustar_dtd: 1 si se desea incrustar el contenido del DTD en el
propio documento, pudiendo abrir y validarlo en cualquier momento a
costa de un mayor tamaño, o 0 de lo contrario.

=back

Devuelve: nada.

=cut

sub escribir_xml
{
    my ($this, %args) = @_;

    ########## XML::Writer
    my $writer;
    if (exists $args{writer}) {
        $writer = $args{writer};
    } else {
        my $opt_outfile = $args{destino_xml};

        $writer = new XML::Writer(ENCODING => "utf-8",
                                  DATA_MODE => 1,
                                  DATA_INDENT => 2,
                                  defined($opt_outfile) ? (OUTPUT => $opt_outfile) : ());
    }

    ########## Declaración XML y DOCTYPE
    $writer->xmlDecl();

    if ($args{incrustar_dtd}) {
        # Generamos DOCTYPE con DTD inline nosotros mismos
        # (XML::Writer sólo imprime públicos y de sistema)
        my $out = $writer->getOutput();
        print $out "<!DOCTYPE script [" . _volcar_fichero($path_to_dtd) . "]>";
    }
    else {
        # Enlazamos con $dtd_fichero del directorio de ficheros compartidos
        $writer->doctype(ACL2::Tags::MAIN_TAG, undef, $path_to_dtd);
    }

    ########## Cuerpo

    # Tag inicial
    $writer->startTag(ACL2::Tags::MAIN_TAG,
                      "acl_version"  => $this->{version},
                      "initOpts"     => $this->{init_opts},
                      "source"       => $this->{ruta_lisp},
                      "source_proof" => $this->{ruta_dem});

    # Añadimos la información de dependencias
    $this->{deps}->escribir_xml($writer);

    # Dump each child as XML
    foreach my $child (@{$this->{children}}) {
        $child->escribir_xml($writer);
    }

    ########## Fin
    $writer->endTag(ACL2::Tags::MAIN_TAG);
    $writer->end();
}

##################################################
#
# MÉTODOS PRIVADOS
#
##################################################

sub _inicializar
{
    my ($this, $writer) = @_;

    my @packages = ();
    my @outputs  = ();
    $this->_prepare_outputs($this->{ruta_dem}, \@packages, \@outputs);

    print STDERR $lh->maketext("Parsing decorated proof script [_1]...",
                               $this->{ruta_tmp_lisp}) . "\n";
    my @program_tokens = $this->_prepare_program($this->{ruta_tmp_lisp});

    # Global state table for this conversion run (useful for things
    # like :dir :system, ADD-INCLUDE-BOOK, and so on)
    my $state = new ACL2::State();

    # Loop over all nodes in the program, skipping whitespace and
    # interleaving comments in program order
    $this->{children}   = [];
    while (@program_tokens > 0) {
        my ($node) = shift @program_tokens;

        for ($node->type) {
            /whitespace/ && next;
            /comment/    && do {
                push @{$this->{children}}, $node;
                next;
            };

            my ($output, $pkg) = (shift @outputs, shift @packages);
            my $orden = ACL2::Orden::Fabrica->crear($node, $output,
                                                    $pkg, $state);
            push @{$this->{children}}, $orden;
        }
    }
}

sub _prepare_outputs {
    my ($this, $path_out, $rv_packages, $rv_outputs) = @_;

    # Load proof script and ACL2 output from disk
    my $salidaACL2 = _volcar_fichero($path_out)
        or croak $lh->maketext("No se pudo abrir [_1]: [_2]",
                               $path_out, $@);

    # Split ACL2's output using prompts
    my @arbol_salidas = split /
         (?:^|(?<=>))  # Start of line, or right after another prompt
         (\S*?) \s?    # Package name (if any)
         (?:\[RAW LISP\]|[pP!s]*)    # raw Lisp indicators and others
         (>+)           # No nested REPLs at this stage
    /xm, $salidaACL2;

    # Process the ACL2 greeting
    my ($acl2_start_msg) = shift @arbol_salidas;
    $acl2_start_msg  =~ /ACL2 \s Version \s ([\.0-9]+)/xs
        or die $lh->maketext("Se esperaba la versión de ACL2");
    $this->{version} = $1;
    $acl2_start_msg  =~ /Initialized \s with \s (\(.+?\))/xs
        or die $lh->maketext("Se esperaban las opciones de ACL2");
    $this->{init_opts} = $1;

    # Remove outputs from the commands in the prologue, up to the
    # point where we start the ACL2 REPL. We cannot use a constant
    # element count, as the number of outputs captured might differ
    # between Lisp implementations (prompts in CLisp, CMUCL and GCL
    # are quite different).
    while (scalar @arbol_salidas > 0) {
        my ($elem) = shift @arbol_salidas;
        last if $elem =~ /ACL2 \s Version \s (?:[\.0-9]+)/x;
    }

    # Join outputs from nested REPLs
    while (scalar @arbol_salidas > 0) {
        my ($pkg, $level, $text) = splice @arbol_salidas, 0, 3;
        if (length($level) == 1) {
            push @$rv_packages, $pkg;
            push @$rv_outputs, $text;
        }
        else {
            $rv_outputs->[-1] .= "\n$pkg $level\n$text";
        }
    }
}

sub _prepare_program {
    my ($this, $path_lisp) = @_;

    # Parse the whole Lisp script
    my $guionLisp  = _volcar_fichero($this->{ruta_tmp_lisp})
        or croak $lh->maketext("No se pudo abrir [_1]: [_2]",
                               $this->{ruta_tmp_lisp},
                               $@);
    my $program    = read_lisp_program(\$guionLisp);

    # Remove commands from the prologue
    while (scalar @{$program->value} > 0) {
        my $token = shift @{$program->value};
        last if ($token->source eq ACL2::Proyecto::END_PROLOGUE());
    }

    # Remove commands from the epilogue
    while (scalar @{$program->value} > 0) {
        my $token = pop @{$program->value};
        last if ($token->source eq ACL2::Proyecto::BEGIN_EPILOGUE());
    }

    return @{$program->value};
}

sub _volcar_fichero {
    my ($ruta, $sref) = @_;
    open FICH, "<", $ruta or return undef;
    return join '', <FICH>;
}

1; # End of ACL2::Demostracion
__END__

