package ACL2::Lisp::BaseNode;

=head1 NAME

ACL2::Lisp::BaseNode - base class for all Lisp AST nodes

=head1 FUNCTIONS

This module contains the following functions:

=cut

#
#    Base class for all Lisp AST nodes
#    Copyright (C) 2009 Antonio García Domínguez
#    $Id$
#
#    This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#

use strict;
use warnings;
use utf8;

use ACL2::Localizacion;
my $lh = ACL2::Localizacion->get_handle();

=head2 value($self)

Returns the semantic value stored in this node:

=over 2

=item *

Symbols, numbers, string, whitespace and most others: characters
comprising the element, without escaping and surrounding delimiters.

=item *

Lists and programs: reference to the list of elements contained within.

=item *

Comma and backquote operators: nested element (the symbol x in ",x",
for instance).

=back

=cut

sub value {
    return $_[0]->{value};
}

=head2 source($self)

Returns the original source code which produced this node. If the
original source code was not provided through the constructor, a
suitable reconstruction will be returned.

=cut

sub source {
    return $_[0]->{source};
}

=head2 type($self)

Returns the type of this AST node. If the AST node is in module
ACL2::Lisp::XandY, its type is "xandy".

=cut

sub type {
    return $_[0]->{type};
}

=head2 new($class, $value, [$source])

Static method which provides a default implementation of the
constructor for all AST nodes. Takes the following arguments:

=over 2

=item *

C<value>: semantic value of the new node, as returned by C<source>.

=item *

C<source>: optional. If passed, it will be used as-is as the original
source from which this node was produced. Otherwise, a suitable
reconstruction will be produced.

=back

=cut

sub new {
    my ($class, $value, $source) = @_;

    my @parts    = split /::/, $class;
    my $basename = lc $parts[-1];
    my $rh_obj = {value => $value,
                  type => $basename,
                  source => (defined($source) ? $source : $value)};

    return bless $rh_obj, $class;
}

=head2 clean_tree($self)

Returns a "clean" version of the AST tree in which all comment and
whitespace nodes have been removed. Source code for the remaining AST
nodes is left untouched (source code for programs and lists will still
contain the original code verbatim, with all the comments and
whitespace).

=cut

sub clean_tree {
    # Default implementation for atomic values which do not need cleaning
    return $_[0];
}

=head2 escribir_xml($self, $writer)

Dumps the current node in XML form to the XML::Writer object
C<writer>.

=cut

sub escribir_xml
{
    croak
        $lh->maketext("Invocado método abstracto [_1]", (caller(1))[3]);
}

1;
