package ACL2::Lisp::Parser;

=head1 NAME

ACL2::Lisp::Parser - Revised implementation of Common Lisp (read) in Perl

=head2 SYNOPSIS

  use ACL2::Lisp::Parser qw(read_lisp_token read_lisp_program);
  my $lispCode = "(+ 1 1) x z";

  # Token by token
  my ($token, $next_offset) = read_lisp_token(\$lispCode, 0);
  while ($next_offset < length($lispCode)) {
     ($token, $next_offset) = read_lisp_token(\$lispCode, $next_offset);
  }

  # All tokens in the input (as an ACL2::Lisp::Program AST node)
  my $program = read_lisp_program(\$lispCode);

=head2 NOTE

This parser, unlike the original version of (read), preserves
whitespace and comments, as we need to be able to reproduce the
original Lisp source code as-is.

=cut

#
#    Revised implementation of Common Lisp (read) in Perl
#    Copyright (C) 2009 Antonio García Domínguez
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#

use strict;
use warnings;
use utf8;
use Carp;

# Load all AST node modules
use ACL2::Lisp::Backquote;
use ACL2::Lisp::BaseNode;
use ACL2::Lisp::Character;
use ACL2::Lisp::Comma;
use ACL2::Lisp::Comment;
use ACL2::Lisp::List;
use ACL2::Lisp::Number;
use ACL2::Lisp::Parser;
use ACL2::Lisp::Program;
use ACL2::Lisp::String;
use ACL2::Lisp::Symbol;
use ACL2::Lisp::Whitespace;

use ACL2::Localizacion;
my  $lh = ACL2::Localizacion->get_handle();

use constant {
    FLAG_REPLACEABLE => "r",
    FLAG_NOT_REPLACEABLE => "n",
    SEEK_SET => 0,
    SEEK_CUR => 1,
};

use Exporter 'import';
our @EXPORT_OK = qw(read_lisp_program read_lisp_token try_read_lisp_token
                    is_potential_number
                    is_whitespace is_macro
                    is_nonterminating_macro is_terminating_macro
                    is_single_escape is_multiple_escape
                    is_alpha_lower is_alpha_upper
                    is_digit is_constituent);

# Used after being localized (see "Temporary values via local()" in
# perlsub) to avoid having to pass and return the line number around
# all the functions which read tokens and characters.
our $line_number = -1;

=head1 FUNCTIONS

=head2 read_lisp_program($rs_input)

Reads Lisp tokens using F<read_lisp_token> repeatedly until reaching
the end of the input. Returns the whole Lisp program as an instance of
ACL2::Lisp::Program. Parsing errors are reported through the $@
special variable, as usual in Perl.

=cut

sub read_lisp_program
{
    my ($rs_input) = @_;

    my @nodes = ();
    my ($node, $offset, $current_line) = (undef, 0, 1);
    while ($offset < length($$rs_input)) {
        ($node, $offset, $current_line)
            = read_lisp_token($rs_input, $offset, $current_line);
        push @nodes, $node if defined($node);
    }
    return ACL2::Lisp::Program->new(\@nodes);
}

=head2 read_lisp_token($rs_input, $offset, [$previous_line])

Reads a token from a reference to a string C<rs_input> containing Lisp
expressions, following an algorithm similar to that described in
section 22.1 of:

    Common Lisp the Language, 2nd edition
    by Guy L. Steele, Thinking Machines, Inc.
    Digital Press
    1990 paperbound
    1029 pages
    ISBN 1-55558-041-6  $39.95

It starts parsing on the C<offset>-th character, starting from 0. It
will return a pair C<(ast_node, offset)> with the corresponding AST
node object from the C<ACL2::Lisp::*> hierarchy and the offset of the
next character which has not been read yet. C<offset =
length($$rs_input)> signals the last token in the input.

Optionally, it can take a C<previous_line> argument with the line
number where the token starts. It is only useful for error reporting.

It returns the token read (as an object of the ACL2::Lisp hierarchy,
except for ACL2::Parser), the offset of the first character which is
not part of the token, and its line number.

See t/lisp-parsing.t for some examples of inputs and their expected
outputs. Parsing errors are reported through the $@ special variable
(see t/lisp-parsing-errors.t for the tests which check that errors are
properly repoted).

=cut

sub read_lisp_token {
    my $rs_input = shift;
    croak $lh->maketext("No input given")  if !defined($rs_input);
    my $offset   = shift;
    croak $lh->maketext("No offset given") if !defined($offset);
    local $line_number = shift || 1;

    my $fh;
    open($fh, "<", $rs_input);
    seek($fh, $offset, SEEK_SET);
    my ($matched_text, $node);
    eval {
        ($matched_text, $node) = _initial_dispatch($fh);
    };
    my $error = $@;
    die $lh->maketext("Parsing error at line [_1]: [_2]",
                      $line_number, lcfirst $error)
        if $error;

    if (defined($matched_text)) {
        return ($node, $offset + length($matched_text), $line_number);
    }
    else {
        return (undef, length($$rs_input), $line_number);
    }
}

=head2 try_read_lisp_token($rs_input, $offset)

Reads a token if it can. It can return the following three values:

=over 2

=item *

C<(token, next)>, with the token read and the offset to the first
character that hasn't been read yet.

=item *

C<(undef, length($$rs_input))> if EOF has been reached.

=item *

C<(undef, next)> if a parsing error occurred, where C<next> is the
offset of the next character to that which produced the parsing error.

=back

It is more or less a fail-safe variant of the read_lisp_token
function, if we are not sure whether the input contains a Lisp token
at the specified offset.

=cut

sub try_read_lisp_token {
    my ($rs_input, $offset) = @_;

    # We don't really need the line number here (no error reporting
    # will be done), but the utility functions expect it
    local $line_number = 0;

    my $fh;
    open($fh, "<", $rs_input);
    seek($fh, $offset, SEEK_SET);
    my ($matched_text, $node);
    eval {
        ($matched_text, $node) = _initial_dispatch($fh);
    };

    if ($@) {
        return (undef, tell($fh));
    }
    elsif (defined($matched_text)) {
        return ($node, $offset + length($matched_text));
    }
    else {
        return (undef, length($$rs_input));
    }
}

#################### PRIVATE METHODS ###########################################

sub _initial_dispatch
{
    my ($fh_chars) = @_;

    # STEP 1 CLTL
    while (1) {
        if (!_eof_reached($fh_chars)) {
            my $char = _read_char($fh_chars);

            # STEP 3 CLTL (modified)
            if (is_whitespace($char)) {
                return _whitespace_token($fh_chars, $char);
            }

            # STEP 4 CLTL
            my $macro_fn = is_macro($char);
            if (defined($macro_fn)) {
                my ($matched_text, $node) = &$macro_fn($fh_chars);
                if (defined($matched_text)) {
                    return ($matched_text, $node);
                }
                next;
            }

            # STEP 5 CLTL
            if (is_single_escape($char)) {
                my $next_char = _read_single_escaped_char($fh_chars);
                return _accumulate_token_even_multiple_escapes(
                    $fh_chars, $char . $next_char, $next_char,
                    FLAG_NOT_REPLACEABLE);
            }

            # STEP 6 CLTL
            if (is_multiple_escape($char)) {
                return _accumulate_token_odd_multiple_escapes(
                    $fh_chars, $char, "", "");
            }

            # STEP 7 CLTL
            if (is_constituent($char)) {
                return _accumulate_token_even_multiple_escapes(
                    $fh_chars, $char, $char, FLAG_REPLACEABLE);
            }

            # STEP 2 CLTL
            croak $lh->maketext(
                "Invalid character found while parsing: '[_1]'",
                $char, $line_number);
        }
        else {
            # End of file
            return (undef, undef);
        }
    }
}

sub _whitespace_token
{
    my ($fh_chars, $txt) = @_;

    while (!_eof_reached($fh_chars)) {
        my $char = _read_char($fh_chars);

        if (is_whitespace($char)) {
            $txt .= $char;
        }
        else {
            _unshift_char($fh_chars, $char);
            last;
        }
    }

    return ($txt, ACL2::Lisp::Whitespace->new($txt));
}

sub _accumulate_token_even_multiple_escapes
{
    my ($fh_chars, $acum_src, $acum_const, $replaceable_flags) = @_;

    # STEP 8 CLTL
    while(1) {
        if (_eof_reached($fh_chars)) {
            return _interpret_token($acum_src, $acum_const, $replaceable_flags);
        }

        my $char = _read_char($fh_chars);
        if (is_constituent($char) || is_nonterminating_macro($char)) {
            $acum_src   .= $char;
            $acum_const .= $char;
            $replaceable_flags .= FLAG_REPLACEABLE;
        }
        elsif (is_single_escape($char)) {
            my $next_char = _read_single_escaped_char($fh_chars);
            $acum_src   .= $char . $next_char;
            $acum_const .= $next_char;
            $replaceable_flags .= FLAG_NOT_REPLACEABLE;
        }
        elsif (is_multiple_escape($char)) {
            $acum_src .= $char;
            return _accumulate_token_odd_multiple_escapes(
                $fh_chars, $acum_src, $acum_const, $replaceable_flags);
        }
        elsif (is_terminating_macro($char) || is_whitespace($char)) {
            _unshift_char($fh_chars, $char);
            return _interpret_token($acum_src, $acum_const, $replaceable_flags);
        }
        else {
            croak $lh->maketext(
                "Invalid character found while parsing: '[_1]'",
                $char, $line_number);
        }
    }
}

sub _accumulate_token_odd_multiple_escapes
{
    my ($fh_chars, $acum_src, $acum_const, $replaceable_flags) = @_;

    # STEP 9 CLTL
    while (1) {
        if (_eof_reached($fh_chars)) {
            croak $lh->maketext(
                "End of file while a token is being accumulated and an "
                . "odd number of multiple escapes have been found",
                $line_number);
        }

        my $char = _read_char($fh_chars);
        if (is_constituent($char) ||
            is_macro($char) ||
            is_whitespace($char)) {
            $acum_src          .= $char;
            $acum_const        .= $char;
            $replaceable_flags .= FLAG_NOT_REPLACEABLE;
        }
        elsif (is_single_escape($char)) {
            my $next_char = _read_single_escaped_char($fh_chars);
            $acum_src   .= $char . $next_char;
            $acum_const .= $next_char;
            $replaceable_flags .= FLAG_NOT_REPLACEABLE;
        }
        elsif (is_multiple_escape($char)) {
            $acum_src .= $char;
            return _accumulate_token_even_multiple_escapes(
                $fh_chars, $acum_src, $acum_const, $replaceable_flags);
        }
        else {
            croak $lh->maketext(
                "Invalid character found while parsing: '[_1]'",
                $char, $line_number);
        }
    }
}

sub _interpret_token
{
    my ($token_src, $token_const, $replaceable_flags) = @_;

    # Use the original source for testing whether it is a number or
    # not: multiple and single escapes make the parser consider the
    # following character as alphabetic, so they inmediately rule out
    # numbers. For instance, (symbolp 12) is NIL, but (symbolp '1\2)
    # is T.
    if (is_potential_number($token_src)) {
        return ($token_src, ACL2::Lisp::Number->new($token_src));
    }
    else {
        # Perform case conversion using the replaceable flag string
        # (emulating the behaviour of readtable-case), but keep the
        # original source as is.
        my $final_const = "";
        foreach my $i (0 .. length($token_const) - 1) {
            my $char      = substr $token_const, $i, 1;
            my $flag      = substr $replaceable_flags, $i, 1;
            $final_const .= $flag eq FLAG_REPLACEABLE ? uc $char : $char;
        }
        return ($token_src,
                ACL2::Lisp::Symbol->new($final_const, $token_src));
    }
}

#################### MACROS ####################################################

sub _macro_string
{
    my ($fh_chars) = @_;

    my $src   = '"';
    my $value = '';
    while (!_eof_reached($fh_chars)) {
        my $char = _read_char($fh_chars);
        if ($char eq '"') {
            $src .= $char;
            return ($src, ACL2::Lisp::String->new($value));
        }
        elsif (is_single_escape($char)) {
            my $next_char = _read_single_escaped_char($fh_chars);
            $src   .= $char . $next_char;
            $value .= $next_char;
        }
        else {
            $src   .= $char;
            $value .= $char;
        }
    }

    croak $lh->maketext(
        "EOF found while looking for the closing quote of a string",
        $line_number);
}

# Flag which tells other functions whether we're inside a backquote
# template
our $inside_backquote = 0;
sub _macro_backquote
{
    my ($fh_chars) = @_;
    local $inside_backquote = 1;
    my ($curr_text, $curr_node) = _initial_dispatch($fh_chars);
    return ("`$curr_text", ACL2::Lisp::Backquote->new($curr_node));
}

sub _macro_sharp
{
    my ($fh_chars) = @_;

    # Read the digit string (if any)
    my $digits = "";
    while (!_eof_reached($fh_chars) && is_digit(_peek_char($fh_chars))) {
        my $char = _read_char($fh_chars);
        $digits .= $char;
    }

    # Read the macro identifier
    if (_eof_reached($fh_chars)) {
        croak $lh->maketext(
            "EOF found while looking for the "
            . "dispatching macro identifier",
            $line_number);
    }
    my $original_id = _read_char($fh_chars);
    # Dispatching macro identifiers are case-insensitive
    my $lc_id       = lc $original_id;

    # Dispatch according to a predefined table
    # TODO: implement more dispatching macros
    my %dispatch_table = (
        "|" => \&_dispatch_block_comment,
        );
    die $lh->maketext(
        "Unknown dispatching macro: '[_1]'",
        "#${digits}${lc_id}",
        $line_number)
        if !exists $dispatch_table{$lc_id};
    my $fn = $dispatch_table{$lc_id};

    my ($matched_text, $node) = &$fn($fh_chars, $digits);
    return ("#" . $digits . $original_id . $matched_text, $node);
}

sub _macro_quote
{
    my ($fh_chars) = @_;

    my ($curr_text, $curr_node) = _initial_dispatch($fh_chars);
    my $matched_text = "'$curr_text";
    return ($matched_text, ACL2::Lisp::List->new(
                [ACL2::Lisp::Symbol->new("QUOTE"),
                 $curr_node],
                $matched_text));
}

sub _macro_lparen
{
    my ($fh_chars) = @_;

    my $matched_text = "(";
    my $rv_tokens    = [];
    while (!_eof_reached($fh_chars) && _peek_char($fh_chars) ne ")") {
        my ($curr_text, $curr_node) = _initial_dispatch($fh_chars);
        $matched_text  .= $curr_text;
        push @$rv_tokens, $curr_node;
    }

    if (!_eof_reached($fh_chars)) {
        # Remove the closing parenthesis from the input
        $matched_text .= _read_char($fh_chars);
        return ($matched_text, ACL2::Lisp::List->new($rv_tokens,
                                                     $matched_text));
    }
    else {
        croak $lh->maketext(
            "EOF found while looking for closing parenthesis "
            . "of a list", $line_number);
    }
}

sub _macro_rparen
{
    croak $lh->maketext("Stray right parent found in input",
                        $line_number);
}

sub _macro_comma
{
    my ($fh_chars) = @_;

    croak $lh->maketext("Comma outside any backquoting was found",
                        $line_number)
        unless $inside_backquote;

    my ($curr_text, $curr_node) = _initial_dispatch($fh_chars);
    return (",$curr_text", ACL2::Lisp::Comma->new($curr_node));
}

sub _macro_colon
{
    my ($fh_chars) = @_;

    my $matched_text = ";";
    while (!_eof_reached($fh_chars) && _peek_char($fh_chars) ne "\n") {
        $matched_text .= _read_char($fh_chars);
    }

    return ($matched_text, ACL2::Lisp::Comment->new($matched_text));
}

#################### DISPATCHING MACROS ########################################

sub _dispatch_block_comment
{
    my ($fh_chars, $digits) = @_;

    # Validate input
    if (length($digits) > 0) {
        croak $lh->maketext(
            "[_1] does not take any integer arguments",
            "#|",
            $line_number);
    }

    # Block comments can be nested (unlike those from PL/I and C)
    my $depth        = 1;

    my $matched_text = "";
    while ($depth > 0 && !_eof_reached($fh_chars)) {
        my $char = _read_char($fh_chars);

        if ($char eq "|") {
            my $next_char = _read_char($fh_chars);
            $matched_text .= $char . $next_char;

            if ($next_char eq "#") {
                --$depth;
            }
        }
        elsif ($char eq "#") {
            my $next_char = _read_char($fh_chars);
            $matched_text .= $char . $next_char;
            if ($next_char eq "|") {
                ++$depth;
            }
        }
        else {
            $matched_text .= $char;
        }
    }
    if ($depth > 0) {
        croak $lh->maketext(
            "EOF while looking for the end of a block comment "
            . "(nested [quant,_1,level] deep)",
            $depth, $line_number);
    }

    return ($matched_text, ACL2::Lisp::Comment->new("#|" . $matched_text));
}

#################### UTILITY FUNCTIONS #########################################

sub _read_single_escaped_char
{
    my ($fh_chars) = @_;
    my $char = _read_char($fh_chars);
    croak $lh->maketext("Found single escape at EOF: expected "
                        . "escaped char",
                        $line_number)
        unless defined($char);
    return $char;
}

sub _eof_reached
{
    return eof($_[0]);
}

sub _read_char
{
    my $c = getc($_[0]);
    ++$line_number if $c eq "\n";
    return $c;
}

sub _peek_char
{
    my ($fh_chars) = @_;
    my $char = _read_char($fh_chars);
    _unshift_char($fh_chars, $char);
    return $char;
}

sub _unshift_char
{
    my ($fh, $c) = @_;
    --$line_number if $c eq "\n";
    seek($fh, -1, SEEK_CUR); # move backwards 1 byte
}

#################### NUMBER IDENTIFICATION #####################################

=head2 is_potential_number($str)

Returns 1 when C<str> constitutes a potential Lisp number, 0 otherwise.

=cut

my $number_token_re = qr!^ # must match the whole string
 [+-]?                     # optional sign
 (?:
   [0-9]+ \.?              # integer
 | [0-9]+ / [0-9]+         # ratio
 | [0-9]* \. [0-9]+        (?:[esfdl] [+-]? [0-9]+)? # floating point
 | [0-9]+ (?: \. [0-9]*)?  (?:[esfdl] [+-]? [0-9]+)  # floating point
 )
 $                         # must match the whole string
!oix;

sub is_potential_number
{
    return $_[0] =~ $number_token_re;
}

#################### CHARACTER IDENTIFICATION ##################################

=head2 is_whitespace($char)

Returns 1 when C<char> is a whitespace character, according to the
Lisp language specification. Otherwise, it returns 0.

=cut

sub is_whitespace
{
    return (index "\015\012\t\f\b ", $_[0]) != -1;
}

=head2 is_macro($char)

Returns 1 when C<char> is a macro character, according to the Lisp
language specification. Otherwise, it returns 0.

=cut

sub is_macro
{
    my %func_map = ('"', \&_macro_string,
                    '`', \&_macro_backquote,
                    '#', \&_macro_sharp,
                    "'", \&_macro_quote,
                    '(', \&_macro_lparen,
                    ')', \&_macro_rparen,
                    ',', \&_macro_comma,
                    ';', \&_macro_colon);

    return $func_map{$_[0]};
}

=head2 is_nonterminating_macro($char)

Returns 1 when C<char> is a non-terminating macro character, according
to the Lisp language specification. Otherwise, it returns 0.

=cut

sub is_nonterminating_macro
{
    return $_[0] eq '#';
}

=head2 is_terminating_macro($char)

Returns 1 when C<char> is a terminating macro character, according to
the Lisp language specification. Otherwise, it returns 0.

=cut

sub is_terminating_macro
{
    return is_macro($_[0]) && !is_nonterminating_macro($_[0]);
}

=head2 is_single_escape($char)

Returns 1 when C<char> is a single escape character, according to the
Lisp language specification. Otherwise, it returns 0.

=cut

sub is_single_escape
{
    return $_[0] eq '\\';
}

=head2 is_multiple_escape($char)

Returns 1 when C<char> is a multiple escape character, according to
the Lisp language specification. Otherwise, it returns 0.

=cut

sub is_multiple_escape
{
    return $_[0] eq "|";
}

=head2 is_digit($char)

Returns 1 when C<char> is a digit, according to the Lisp language
specification. Otherwise, it returns 0.

=cut

sub is_digit
{
    return $_[0] ge "0" && $_[0] le "9";
}

=head2 is_alpha_lower($char)

Returns 1 when C<char> is an lowercase alphabetic character, according
to the Lisp language specification. Otherwise, it returns 0.

=cut

sub is_alpha_lower
{
    return $_[0] ge "a" && $_[0] le "z";
}

=head2 is_alpha_upper($char)

Returns 1 when C<char> is an uppercase alphabetic character, according
to the Lisp language specification. Otherwise, it returns 0.

=cut

sub is_alpha_upper
{
    return $_[0] ge "A" && $_[0] le "Z";
}

=head2 is_constituent($char)

Returns 1 when C<char> is a constituent, according to the Lisp
language specification. Otherwise, it returns 0.

=cut

sub is_constituent
{
    return is_alpha_lower($_[0])
        || is_alpha_upper($_[0])
        || is_digit($_[0])
        || (index "@!\$%&*+-./:<=>?[]\b\x7f^_{}~", $_[0]) != -1;
}

1;

